/**
 * This file Copyright (c) 2018 Magnolia International
 * Ltd.  (http://www.magnolia-cms.com). All rights reserved.
 *
 *
 * This file is dual-licensed under both the Magnolia
 * Network Agreement and the GNU General Public License.
 * You may elect to use one or the other of these licenses.
 *
 * This file is distributed in the hope that it will be
 * useful, but AS-IS and WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE, TITLE, or NONINFRINGEMENT.
 * Redistribution, except as permitted by whichever of the GPL
 * or MNA you select, is prohibited.
 *
 * 1. For the GPL license (GPL), you can redistribute and/or
 * modify this file under the terms of the GNU General
 * Public License, Version 3, as published by the Free Software
 * Foundation.  You should have received a copy of the GNU
 * General Public License, Version 3 along with this program;
 * if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * 2. For the Magnolia Network Agreement (MNA), this file
 * and the accompanying materials are made available under the
 * terms of the MNA which accompanies this distribution, and
 * is available at http://www.magnolia-cms.com/mna.html
 *
 * Any modifications to this file must keep this entire header
 * intact.
 *
 */
package info.magnolia.ui.vaadin.form.field;

import info.magnolia.icons.MagnoliaIcons;
import info.magnolia.ui.vaadin.gwt.client.form.field.FieldLayoutState;

import java.util.function.Consumer;

import org.apache.commons.lang3.StringUtils;
import org.vaadin.jonatan.contexthelp.ContextHelp;

import com.google.common.base.Preconditions;
import com.vaadin.shared.ui.ContentMode;
import com.vaadin.ui.AbstractField;
import com.vaadin.ui.Button;
import com.vaadin.ui.Component;
import com.vaadin.ui.HorizontalLayout;
import com.vaadin.ui.Label;
import com.vaadin.ui.TabSheet;
import com.vaadin.ui.UI;
import com.vaadin.ui.VerticalLayout;

/**
 * A field layout for a field, supporting displaying the error message.
 */
public class FieldLayout extends VerticalLayout implements Component.Focusable {

    private final ContextHelp contextHelp;
    private final boolean isRequired;

    private Label statusLabel;
    private final Component field;
    private final String description;

    public static FieldLayout of(Component field, String description) {
       return of(field, description, false);
    }

    public static FieldLayout of(Component field, String description, boolean isRequired) {
        return new FieldLayout(field, description, isRequired);
    }

    private FieldLayout(Component field, String description, boolean isRequired) {
        this.isRequired = isRequired;
        Preconditions.checkNotNull(field);
        this.field = field;
        this.description = description;
        this.contextHelp = UI.getCurrent().getExtensions().stream()
                .filter(ContextHelp.class::isInstance)
                .findFirst()
                .map(ContextHelp.class::cast)
                .orElse(null);
        initContent();
    }

    private void initContent() {
        setMargin(false);
        setVisible(field.isVisible());

        statusLabel = new Label(null, ContentMode.HTML);
        statusLabel.setVisible(false);
        statusLabel.addStyleName("validation-display");

        Component horizontalLayout = field;
        if (contextHelp != null && hasTranslatedDescription()) {
            final Button helpButton = new Button(MagnoliaIcons.HELP_MARK);
            helpButton.addStyleName("help-button v-button-icon icon-button");

            contextHelp.extend(UI.getCurrent());
            contextHelp.addHelpForComponent(helpButton, description);

            helpButton.addClickListener(event -> {
                contextHelp.showHelpFor(helpButton);
            });
            HorizontalLayout descriptionLayout = new HorizontalLayout(field, helpButton);
            descriptionLayout.addStyleName("help-button-field-layout");
            descriptionLayout.setSizeFull();
            descriptionLayout.setExpandRatio(field, 1);
            descriptionLayout.setExpandRatio(helpButton, 0);
            descriptionLayout.setSpacing(false);
            horizontalLayout = descriptionLayout;
        }

        setCaption(field.getCaption());
        setRequiredIndicatorVisible(isRequired);

        addComponents(horizontalLayout, statusLabel);
    }

    @Override
    public void setCaption(String caption) {
        super.setCaption(caption);
    }

    @Override
    public void setRequiredIndicatorVisible(boolean visible) {
        if (field instanceof AbstractField) {
            ((AbstractField) field).setRequiredIndicatorVisible(false);
        }
        getState().required = visible;
    }

    @Override
    protected boolean isRequiredIndicatorVisible() {
        return getState().required;
    }

    @Override
    protected FieldLayoutState getState() {
        return (FieldLayoutState) super.getState();
    }

    @Override
    public void focus() {
        performTabSelection(field);
        if (field instanceof Focusable) {
            ((Focusable) field).focus();
        }
    }

    @Override
    public int getTabIndex() {
        return 0;
    }

    @Override
    public void setTabIndex(int tabIndex) {
    }

    public Consumer<String> getValidationStatusHandler() {
        return errorMessage -> {
            statusLabel.setValue(errorMessage);
            if (errorMessage.isEmpty()) {
                field.removeStyleName("error");
                statusLabel.setVisible(false);
            } else {
                field.addStyleName("error");
                // Only show the label when validation has failed
                statusLabel.setVisible(true);
            }
        };
    }

    public Component getField() {
        return field;
    }

    private boolean hasTranslatedDescription() {
        return StringUtils.isNotEmpty(description) && (StringUtils.endsWith(description, ".") || !StringUtils.contains(description, ".") || StringUtils.contains(description, " "));
    }

    /*
     * When focusing field within multiple tabs form we reach TabSheet instance among parents of field
     */
    private void performTabSelection(Component f) {
        Component t = f;
        do {
            f = t;
            t = f.getParent();
        } while (!(t == null || t instanceof TabSheet));
        if (t != null) {
            ((TabSheet) t).setSelectedTab(f);
        }
    }
}